@@ -1,5 +1,5 @@
 /*
-    Copyright (c) 2005-2021 Intel Corporation
+    Copyright (c) 2005-2024 Intel Corporation
 
     Licensed under the Apache License, Version 2.0 (the "License");
     you may not use this file except in compliance with the License.
@@ -221,7 +221,7 @@ void print_my_value(MyKeySecond<K, V> const &i) {
 
 template<>
 void print_my_value(std::string const &i) {
-    INFO("\"" << i.c_str() << "\"" );
+    INFO("\"" << i << "\"" );
 }
 
 //
@@ -245,10 +245,10 @@ struct my_struct_key<K&, V> {
     }
 };
 
-using tbb::detail::d1::type_to_key_function_body;
-using tbb::detail::d1::hash_buffer;
+using tbb::detail::d2::type_to_key_function_body;
+using tbb::detail::d2::type_to_key_function_body_leaf;
+using tbb::detail::d2::hash_buffer;
 using tbb::detail::d1::tbb_hash_compare;
-using tbb::detail::d1::type_to_key_function_body_leaf;
 
 template<class K, class V> struct VtoKFB {
     typedef type_to_key_function_body<V, K> type;
@@ -1172,7 +1172,7 @@ public:
         INFO("\n");
         for(int i = 0; i < MaxPorts; ++i) {
             for(int j = 0; j < MaxNInputs; ++j) {
-                all_input_nodes[i][j] = NULL;
+                all_input_nodes[i][j] = nullptr;
             }
         }
         for(int nInputs = 1; nInputs<=MaxNInputs; ++nInputs) {
@@ -1199,7 +1199,7 @@ public:
             reset_outputCheck(TUPLE_SIZE, Count);
 
             for(int i = 0; i < Count; i++) {
-                CHECK_MESSAGE(outq2.try_get(v), "");;
+                CHECK_MESSAGE(outq2.try_get(v), "");
                 input_node_helper<TUPLE_SIZE, JType>::check_value(i, v, not_out_of_order);
             }
             check_outputCheck(TUPLE_SIZE, Count);
